'''
Created on Jan 24, 2020

@author: vladislavkargin

Various tools needed for manipulating Trees and Maps

'''
import copy
import numpy as np
from scipy.integrate import quad
import matplotlib.pyplot as plt

import rgs.pmaps.OrderedTree as ot
#import pmaps.PlanarGraph as pg
import rgs.pmaps.Triangulation as trngl

 
def reweight(w, rho):
    '''
    Reweights a weight sequence by \rho. 
    For an array w, sends w_k to w_k*rho^k and divides the resulting sequence 
    by its sum, so the result is a probability sequence. Returns 
    the reweighted sequence.
    returns a reweighted sequence and its mean. 
    
    The argument w should be a numpy 1-dim array
    '''  
    rw = w
    for i in range(len(w)):
        rw[i] = w[i] * rho ** i         
    #normalizing sum
    s = np.sum(rw)
    rw = rw / s
    mu = 0.
    for i in range(len(rw)):
        mu = mu + i * rw[i]
    return rw, mu


def rescale(w):
    '''
    Obtains a probability weights with expectation 1 from 
    an arbitrary sequence of (non-negative) weights.
    Returns the rescaled sequence
    The argument should be an object convertible to a numpy 1dim array
    '''
    pp, mu = reweight(np.array(w), 1.)
    #initializing search window [a b] for the re-weight parameter.
    b = 1.;
    delta = 0.05;
    while (abs(mu - 1.) > 10 ** (-10)):
        if (mu < 1.):
            b = 1 + delta;
            pp, mu = reweight(pp, b)
            while (mu < 1.):
                pp, mu = reweight(pp, b)
        else:
            b = 1 - delta;
            pp, mu = reweight(pp, b)
            while (mu > 1.):
                pp, mu = reweight(pp, b)
        delta = delta/2;
    return pp

def alphaShift(alpha, w):
    '''
    using a parameter alpha and a weight sequence w
    uses an exponential rescaling on w1, w2, ...  and 
    finds a new sequence wn such that wn[0] = alpha and 
    w1 a probability sequence with mean 1
    '''
    w1 = np.concatenate(([0],w[1:])) #the sequence w with 0 instead of w_0
    #print("w1 = ", w1)
    pp, mu = reweight(w1, 1.)
    pp = pp * (1 - alpha)
    #print("alpha = ", alpha)
    #print("pp = ", pp)
    mu = mu *  (1 - alpha)
    #initializing search window [a b] for the re-weight parameter.
    b = 1.;
    delta = 0.05;
    while (abs(mu - 1.) > 10 ** (-10)):
        if (mu < 1.):
            b = 1 + delta;
            pp, mu = reweight(pp, b)
            pp = pp * (1 - alpha)
            mu = mu *  (1 - alpha)
            while (mu < 1.):
                pp, mu = reweight(pp, b)
                pp = pp * (1 - alpha)
                mu = mu *  (1 - alpha)
        else:
            b = 1 - delta;
            pp, mu = reweight(pp, b)
            pp = pp * (1 - alpha)
            mu = mu *  (1 - alpha)
            while (mu > 1.):
                pp, mu = reweight(pp, b)
                pp = pp * (1 - alpha)
                mu = mu *  (1 - alpha)
        delta = delta/2;
    pp[0] = alpha    
    return pp

def treeToDyckPath(tree):
    '''converts a planar rooted tree to a Dyck path'''
    A = [-1]* tree.graph.V 
    dyckPath = []
    v = tree.graph.root
    stack = [v]
    A[v] = 0 
    #dfsTimes = [0]  #the list of times when a new vertex was discovered.
    n = 1
    while (len(stack) > 0):
        #print("stack = " + str(stack));
        v = stack[-1]; #peaking at the top of the stack
        #print("processing v = " + str(v))
        flag = False #can we find unprocessed neighbors?
        d = tree.graph.degree(v)
        for i in range(d):
            w = tree.graph.adj[v][i]
            if (A[w] < 0): #this neighbor has not been explored previously
                A[w] = n
                n = n + 1
                stack.append(w)
                #print(dfsPath)
                dyckPath.append(1)
                #dfsTimes.append[len(dyckPath)]
                flag = True
                break #stop searching for an uexplored neighbor
            else: 
                continue
        if (not flag): 
            stack.pop() #no unexplored neighbor around v. Remove it from consideration.
            dyckPath.append(-1)
    del dyckPath[-1]
    return  dyckPath   

def treeToLukasPath(tree):
    '''converts a planar rooted tree to the corresponding Lukasiewicz path
    Assumes that the vertices are already named in the DFS order'''
    path = []
    for v in range(tree.graph.V):
        path.append(tree.outDegree(v) - 1)
    return path    

def treeToHeightPath(tree):
    '''converts a planar rooted tree to the corresponding height path
    Assumes that the vertices are already named in the DFS order'''
    path = []
    for v in range(tree.graph.V):
        path.append(tree.heightVertex(v))
    return path 
        
def cyclicShift(path):
    '''cyclic shift of the path: it finds the first minimum of the path and rotate 
    the sequence of steps so that it starts from this minimum. It is useful for 
    creating Dyck and Lucasiewicz paths. 
    arguments: path (in fact, it is a sequence of steps)
    returns: the shifted path
    '''
    m = 0
    pos = 0
    S = 0
    n = len(path)
    for i in range(n):
        S = S + path[i]
        if (S < m):
            m = S
            pos = i + 1
    if (pos == n):
        return path
    else:
        path = np.concatenate([path[pos:n], path[0:pos]])
        return path.astype(np.int64)   

def randomBridge(n, w, ITER = 10000, SEED = 0):
    ''' creates a random walk bridge that starts at 0, ends at -1 and have
    the distribution of steps w. (w_0 corresponds to step -1)
    returns: the path of the bridge as a sequence of steps. 
    '''
    K = len(w);
    pp = rescale(w)
    #get a sample from multinomial distribution
    count = 0
    mu = 0
    if (SEED != 0):
        np.random.seed(SEED)
    else:
        np.random.seed()
    while (mu != -1 and count < ITER):
        #repeat until the required property of the path is 
        #ensured or too many iteration are used. 
        X = np.random.multinomial(n, pp)
        mu = 0
        for i in range(K):
            mu = mu + (i - 1) * X[i]
        count = count + 1

    if (count == ITER and mu != -1):
        print("RandomLukasiewicz says: Could not find a path that would "
                    + "end in -1 after " + str(ITER) + "iterations")

    #convert multinomial array X to a path
    path = np.array([]);
    for i in range(K): 
        path = np.concatenate([path, np.array([i - 1] * X[i])])
    #print(path)    
    np.random.shuffle(path)
    return path

    '''
     * generate a random L􏰀ukasiewicz path that have n steps, starts at
      (0,0)  and never goes below the horizontal line except at the 
      last step and stops at (n,-1).
     * 
     * A L􏰀ukasiewicz path is a path of a random walk that can steps of 
     size in the set {-1, 0, 1, 2, 3, ...} and that do not go below 
     the horizontal line except at the last step. 
     * 
     * The weights of steps are given in the array w. So w0 is the 
     * probability of the -1 step, w1 probability of 0, w2 probability of 1
     * and so on. 
     * 
     * The generator works in 3 steps
     * 0) from the given sequence of weights generate a sequence of probabilities
     * such that the expectation is close to 1.
     * 1) generate a multinomial partition of n according do the 
     * distribution pp. So we will get N0, N1, N2, ... which will sum
     * to n and have probabilities p0, p1, p2 and so on. 
     * The number Nj represent the number of times the step j-1 occurs in 
     * the path. 
       2) repeat until \sum (j - 1)Nj hits -1 
     * 2) then we create a array of n steps and randomly permute it. 
     * 3) then we do a Vervaat transform to make sure that the path
     * stays above the horizontal line. 
     * 4) repeat until we hit -1
     * 
     * 
     * @param n length of the path
     * @param w weights of the steps. 
     * @param ITER maximum number of attempts to generate the path 
     * @param SEED if seed not 0 then uses provided seed to generate random sequence
     * @return
    '''
def randomLukasiewicz(n, w, ITER = 10000, SEED = 0):
    #generate a random L􏰀ukasiewicz path that have n steps, starts at
    #(0,0)  and never goes below the horizontal line except at the 
    #last step and stops at (n,-1).
    
    path = randomBridge(n, w, ITER = ITER, SEED = SEED)
    shifted_path = cyclicShift(path)
    return shifted_path
        
def randomLukasSlim(k, n, w, ITER = 10000, SEED = 0):
    '''
    This is a slight generalization of randomLukasiewicz function that we need to 
    generate slim trees. The generalization is that we 
    1) build a sequence of n i.i.d variables that sum to -1 
    using the alpha shifted distribution and making sure that there are k steps of size - 1
    2) proceeding as usual in random Lukasiewicz
    arguments: k - number of -1 
    n - length of the sequence 
    w - weight of increments.  
    ITER number of attempts to find the sequence
    SEED - for replication purposes. 
    '''
    weights_len = len(w);
    alpha = k/n
    pp = rescale(w)
    #print('pp = ', pp)
    wn = alphaShift(alpha, pp)
    #get a sample from multinomial distribution
    count = 0
    mu = 0
    if (SEED != 0):
        np.random.seed(SEED)
    while (mu != -1 or count < ITER):
        #print(count)
        #repeat until the required property of the path is 
        #ensured or too many iteration are used. 
        #print('wn = ', wn)
        X = np.random.multinomial(n, wn)
        #print('X = ', X)
        if X[0] != k:
            count = count + 1
            continue
        mu = 0
        for i in range(weights_len):
            mu = mu + (i - 1) * X[i]
        #print("mu =" + str(mu))
        count = count + 1

    if (count == ITER and mu != -1):
        print("Random LukasSlim says: Could not find a path that would "
                + "end in -1 after " + ITER + "iterations")
            #convert multinomial array X to a path
    path = np.array([]);
    for i in range(weights_len): 
        path = np.concatenate([path, np.array([i - 1] * X[i])])
    #print(path)    
    np.random.shuffle(path)        
    #find the minimum 
    m = 0
    pos = 0
    S = 0
    for i in range(n):
        S = S + path[i]
        if (S < m):
            m = S
            pos = i + 1
    if (pos == n):
        return path
    else:
        path = np.concatenate([path[pos:n], path[0:pos]])
        return path.astype(np.int64)  


        
def randomLukas31(n, ITER = 10000, SEED = 0):  
    '''
 * This is a variant of the method that generate Lukasiewicz paths 
 * that will be used to generate random tiangulations according to 
 * Poulalhon - Schaeffer method 
 * We need a path that consists of steps 3 and -1 such that it has 4n - 2 elements 
 * and total weight (sum of steps) -2.  The path should be always above -2 except at
 * the last step.  
 *  
    '''    
    #length of the path 
    N = 4 * n - 2
    pp = [0.75, 0., 0., 0., 0.25]
    K = len(pp) #K = 5
    #get a sample from multinomial distribution
    count = 0 #number of iterations
    mu = 0    #value at the end of the path
    if (SEED != 0):
        np.random.seed(SEED)
    while (mu != -2 or count < ITER):
        #repeat trials until the required property of the path is 
        #ensured or too many iteration are used. 
        X = np.random.multinomial(N, pp)
        mu = 0
        for i in range(K):
            mu = mu + (i - 1) * X[i]
        count = count + 1
    if (count == ITER and mu != -2):
            print("Could not find a path that would "
                    + "end in -2 after " + ITER + "iterations")
    #convert multinomial array X to a path
    path = np.array([]);
    for i in range(K): 
        path = np.concatenate([path, np.array([i - 1] * X[i])])
        #print(path)    
    np.random.shuffle(path)
    #plt.plot(np.cumsum(path))
    #find the minimum 
    m = 0
    pos = 0
    S = 0
    for i in range(N):
        S = S + path[i]
        if (S < m):
            m = S
            pos = i + 1
    #print("pos = " + str(pos))
    if (pos == N):
        return path
    else:
        path = np.concatenate([path[pos:N], path[0:pos]])
        return path.astype(np.int64)
    
              
def decodeToPSTree(path):
    ''' 
    This will take a path with steps 3 and - 1
    (typically generated by randomLukas31) and convert it to 
    a tree where each inner node has only 2 leafs. 
    arguments: path
    returns: the tree
    '''  
    n = (len(path) + 2) /4
    #print("n = " + str(n))
    tree = ot.OrderedTree(3 * int(n))
    #in a sense we trying to do dfs. I believe the path will always start with 3
    # which means we go down from an inner node.
    x = 0 #name of the currently available vertex in the tree
    stack = [x]
    leafMap = {x : 0} #this will be used to check how many leaves vertex x already has
    x = x + 1 #the next available step
    i = 0 #currently processed step in the path
    
    while (i < len(path) and len(stack) > 0):
        #print("stack = " + str(stack))
        #print("i = " + str(i) + "; path[i] = " + str(path[i]))
        v = stack[-1]; #peaking at the top of the stack
        #print("processing vertex v = " + str(v))
        #print("leafMap[v] = " + str(leafMap[v]))
        if (path[i] == 3): #the step is 3 meaning just go down the tree
            tree.graph.addEdge(v, x)
            #print("Adding edge (" + str(v) + ", " + str(x) + ")")
            stack.append(x)
            leafMap[x] = 0
            x = x + 1
            i = i + 1
        else: #the step is - 1
            if (leafMap[v] < 2): #we can add another edge
                tree.graph.addEdge(v, x)
                #print("Adding edge (" + str(v) + ", " + str(x) + ")")
                leafMap[v] = leafMap[v] + 1
                x = x + 1
                i = i + 1 
            else:  #this vertex is done. 
                i  = i + 1
                stack.pop()
                continue
    #print(tree)        
    return tree        

def randomLukasBinary(k, n, ITER = 10000, SEED = 0):
    '''
    generates a random Lukasiewicz path that consist of k steps of size -1, 
    k-1 steps of size 1 and n-2k+1 steps of size 0 
    '''
    #print("seed = " + str(SEED))
    if (SEED != 0):
        np.random.seed(SEED)
    if n < 2 * k + 1:
        print("randomLukasBinary says: n = " + str(n) + " is too small for k = " + str(k))
        return []
    path = [-1] * k + [0] * (n - 2*k + 1) + [1] * (k - 1)
    np.random.shuffle(path)
    m = 0
    pos = 0
    S = 0
    for i in range(n):
        S = S + path[i]
        if (S < m):
            m = S
            pos = i + 1
    #print("pos = " + str(pos))
    if (pos == n):
        return path
    else:
        path = np.concatenate([path[pos:n], path[0:pos]])
        return path.astype(np.int64)
    return path

def decoratedPathTree(n, SEED = 0):
    '''
    builds a (planar) tree that consists of a path of n inner vertices with each inner
    vertex connected to two leaves. It is going to be used in experiments on triangulations
    '''
    if (SEED != 0):
        np.random.seed(SEED)
    tree = ot.OrderedTree(V = 3 * int(n))
    inner = 0
    for i in range(n - 1):
        tree.graph.addEdge(inner, 3 * i + 1)
        tree.graph.addEdge(inner, 3 * i + 2)
        tree.graph.addEdge(inner, 3 * i + 3)
        r = np.random.randint(1,4)
        inner = 3 * i + r
    tree.graph.addEdge(inner, 3 * (n - 1) + 1)
    tree.graph.addEdge(inner, 3 * (n - 1) + 2)
    tree.dfsRename(0)
    return tree
    


def verdiereFunction(x):
    '''
    calculates the integral of atan(e^t) from -\infty to x
    '''
    [v, _] = quad(lambda t: np.arctan(np.exp(t)), -np.inf, x)
    return v



def fromTriangulation(trntn):
    '''
    builds a tree from a triangulation trntn. It is supposed to be the inverse of the 
    map buildTriangulation in PlanarGraph module
    Reference: Poulalhon and Schaeffer "Optimal Coding and Sampling of Triangulations"
    
    Arguments: trntn - a triangulation
    Returns: trFinal - a BT tree (that is a tree where each inner node have two leaves attached.
    '''
    V = trntn.graph.V
    v1 = trntn.graph.root
    v2 = trntn.graph.rootArrow
    boundary = trntn.graph.getOuterBoundary(u0 = v2)
    v0 = boundary[2] 
    T0, T1, T2 = trntn.calcRealizer()
    
    #print("T0 = " + str(T0))
    #print("T1 = " + str(T1))
    #print("T2 = " + str(T2))
      
    tr = copy.deepcopy(trntn)   
    tr.graph.removeEdge(v0, v1)
    tr.graph.removeEdge(v0, v2)
    
    #initialization of the stack
    stack = trntn.graph.adj[v0]
    stack.remove(v1)
    stack.remove(v2)
    #print("Stack = " + str(stack))
    
    #now we need to add two new vertices (and leaves leading to them)
    tr.graph.addVertex()
    tr.graph.addVertex()
    tr.graph.addEdge(v0, V)
    tr.graph.addEdgeBeforeAfter(v0, V + 1, tr.graph.adj[v0][0],0)
    
    #change root and rootArrow
    tr.graph.root = V + 1
    tr.graph.rootArrow = v0
    
    nvertices = V + 2 #current number of vertices in graph tr.  
    marked = [] #processed edges   
    #counter = 0 #this is for debugging purposes
     
    while len(stack) > 0:
        v = stack[-1]
        
        #we are going around the vertex [in the original graph in the clockwise direction]
        # from an edge on the directed path to v2 
        # to an edge on the directed path to v1
        # and do some operations:
        # if we encounter an edge directed to current v, and this edge is not marked (not processed),
        # we mark this endpoint and add it to the stack and proceed 
        # with this new stack element,
        # if we encounter an edge not directed to v, (it must be on a directed path to v2 or v1)
        # and it is not marked, we mark it, remove it in the new graph, add a new leaf in the new graph and 
        #  continue going around v. 
        # when we are finished exploring edges from v between v1 and v2, we remove the vertex from the stack
        # and proceed with a new element on the stack
        
        start = T2[v]
        finish = T1[v]
        spisok = [] #the edges that go from start to finish edges from v in clockwise direction
        #constructing the list spisok
        ind = trntn.graph.adj[v].index(start) 
        x = trntn.graph.adj[v][ind]
        spisok.append(x)
        while x != finish:      
            ind = (ind - 1) % len(trntn.graph.adj[v])
            x = trntn.graph.adj[v][ind]
            spisok.append(x)
        #print("spisok = " + str(spisok))
        #processing edges in spisok
        isNewEdgeFound = False
        for u in spisok:
            if [v, u] in marked:
                continue
            else:
                isNewEdgeFound = True
                marked.append([v, u]) 
                #TODO: edges should be added in clockwise direction - this is different
                #then simply addEdge below
            if u == start or u == finish: 
                tr.graph.removeEdge(v,u)
                tr.graph.addVertex()
                if u == start:
                    tr.graph.addEdge(v, nvertices)
                else:
                    tr.graph.addEdgeBeforeAfter(v, nvertices, tr.graph.adj[v][1], 0) #this behavior is not 
                        #totally in agreement with Poulalhon - Shaeffer paper, but I cannot quite 
                        #understand how the leaves can be at any place but on the ends of the list
                        #of children.
                nvertices = nvertices + 1
            else:
                stack.append(u)
                break
        if not isNewEdgeFound:
            stack.remove(v)
            #counter = counter + 1
            #print("Vertex " + str(v) + "has been processed")
            #print("After vertex " + str(v) + "is processed, the graph is \nç")
            #print(tr)
            #tr.draw(drawLabels = True, calculateRepr = "Leaves", title = "After v = " + str(v) + "is processed")
            #tr.draw(drawLabels = True, title = "After v = " + str(v) + "is processed")
            
        #finally we should remove the edge between v1 and v2 and both vertices themselves
    tr.graph.removeEdge(v1, v2)
    #tr.draw(drawLabels = True, calculateRepr = "Leaves")
    
    #removing v1 and v2 and renaming vertices in dfs order.
    vertices = list(range(tr.graph.V))
    vertices.remove(v1)
    vertices.remove(v2)  
    trFinal = ot.fromGraph(tr.graph.subgraph(vertices, tr.graph.root, tr.graph.rootArrow))    
    return trFinal
    

'''
For testing methods
'''
def main():
    print('Tree Utility is running')
    '''
    Here we check that random Lukasiewicz method works
    '''
    '''
    w = np.array([1, 2, 3])
    rho = 3
    rw, mu = reweight(w, rho)
    print(rw)
    print(mu)
    rw = rescale(w)
    print(rw)
    n = 100
    path = randomLukasiewicz(n, w)
    print(path)
    plt.plot(np.cumsum(path))
    plt.show()
    '''
    
    '''
    Another tree based on Lukasiewicz paths and used to create triangulations
    '''
    '''
    n = 10
    path = randomLukas31(n)
    decodeToPSTree(path)
    plt.plot(np.cumsum(path))
    tree = decodeToPSTree(path)
    print(tree)
    tree.draw(drawLabels = True)
    plt.show()
    '''
    '''
    Test that Verdiere function is working
    '''
    '''
    x = -5
    v = verdiereFunction(x)
    print(v)
    '''
    
    '''
    A binary-unary tree based on Lukasiewicz method.
    '''
    '''
    n = 1000
    k = 300
    path = randomLukasBinary(k, n)
    #print(path)
    plt.plot(np.cumsum(path))
    tree = ot.fromLukasiewicz(path)
    tree.draw()
    '''
    
    '''
    a decorated path tree (path with 2 leaves attached to each vertex)
    '''
    '''
    n = 30
    seed = 123
    tree = decoratedPathTree(n, SEED = seed)
    print(tree)
    tree.draw(drawLabels = False)
    '''
 
    #n = 8
    #seed = 23167
    n = 5
    seed = 0
    #path = [3, -1, -1, -1, 3, -1, -1, -1, -1, -1]
    #btTree = decodeToPSTree(path)
    #trntn = pg.buildTriangulation(btTree)
    trntn, btTree = trngl.randomTriangulation(n, SEED = seed)
    print("btTree is " + str(btTree))
    btTree.draw(drawLabels = True)
    #trntn = pg.randomTriangulation(n, SEED = seed, method = 'decoratedPath')
    _, ax = trntn.draw(drawLabels = True, calculateRepr = "Schnyder")
    realizer = trntn.calcRealizer()
    trntn.drawRealizer(ax, realizer)
    
    
    '''
    tree that we got back from the realizer
    '''
    tr = fromTriangulation(trntn)
    print(tr)
    tr.draw(drawLabels = True, calculateRepr = "Leaves")
    #plt.show()
    
    '''
    TODO something not OK below
    
    newTrntn = pg.buildTriangulation(tr)
    _, ax = newTrntn.draw(drawLabels = True, calculateRepr = "Schnyder")
    realizer = newTrntn.calcRealizer()
    newTrntn.drawRealizer(ax, realizer)
    '''


    '''
    This checks the the generation of the Lukasiewicz path for the slim trees
    
    '''
    w = [1, 1, 1, 1]
    wn = alphaShift(0.3, w)
    print("Shifted weights = " + str(wn) + "\n have sum = " + str(wn.sum()) )
    _, mu = reweight(wn, 1.)
    print(" and expectation " + str(mu))
    
    n = 99
    alpha = 1/3
    k = np.floor(alpha * n)
    print(k)
    
    path1 = randomLukasiewicz(n, w)
    path2 = randomLukasSlim(k, n, w)
    print(path2)
    print(np.sum(path2 == -1))
    
    fig1, (ax1, ax2) = plt.subplots(nrows=1, ncols=2)
    ax1.plot(np.cumsum(path1))
    ax2.plot(np.cumsum(path2))
    ax1.grid(True)
    ax2.grid(True)
    plt.show()
    
    
    
    print('Tree Utility has finished')
    
if __name__ == '__main__':
    main()


'''
Created on Aug 29, 2020

@author: vladislavkargin
'''

''' Various function for SlimTree paper '''

import numpy as np
import matplotlib.pyplot as plt

import rgs.pmaps.OrderedTree as ot
import rgs.pmaps.TreeUtility as tutil

def slimTree():
    pass



'''
For testing methods
'''
def main():
    
    '''
    We plot several processes for a given tree: contour, Lukasiewisz and height
    
    '''
    n = 8
    w = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    #w = [1, 0, 1]
    seed = 123
    _, ((ax1, ax2), (ax3, ax4)) = plt.subplots(nrows=2, ncols=2)
    
    tree = ot.randomTree(n, w, SEED = seed)
    tree.draw(drawLabels = True, ax = ax1)
    print(tree)
    path = tutil.treeToDyckPath(tree)
    DyckPath = np.concatenate(([0], np.cumsum(path)))
    ax2.plot(DyckPath)
    for i in range(len(path)):
        if path[i] == 1:
            ax2.plot(i + 1, DyckPath[i + 1],'ro')
    ax2.grid(True)
    ax2.set_title("Contour")
    ax2.set_xticks(list(range(2*n-1)))
    ax2.set_yticks(list(range(max(DyckPath))))
    
    path = tutil.treeToLukasPath(tree)
    LukasPath = np.concatenate(([0], np.cumsum(path)))
    ax3.plot(LukasPath)
    ax3.grid(True)
    ax3.set_title("Lukasiewics")
    ax3.set_xticks(list(range(n + 1)))
    ax3.set_yticks(list(range(-1, max(LukasPath))))
    
    HeightPath = tutil.treeToHeightPath(tree)
    ax4.plot(HeightPath)
    ax4.grid(True)
    ax4.set_title("Height")
    ax4.set_xticks(list(range(n + 1)))
    ax4.set_yticks(list(range(0, max(HeightPath))))
    
    
    
    
    n = 21
    w = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    seed = 74
    path1 = tutil.randomBridge(n, w, SEED = seed)
    print(path1)
    walk1 = np.concatenate(([0],np.cumsum(path1)))
    print(walk1)
    walk2 = np.concatenate(([0],np.cumsum(tutil.cyclicShift(path1))))
    print(walk2)
    
    
    _, (ax1, ax2) = plt.subplots(nrows=1, ncols=2)
    ax1.plot(walk1)
    ax2.plot(walk2)
    ax1.grid(True)
    ax1.set_xticks(list(range(0,n + 1,2)))
    ax2.grid(True)
    ax2.set_xticks(list(range(0,n + 1,2)))
    ax2.set_yticks(list(range(-1, max(walk2))))
    
    plt.show()
    pass

if __name__ == '__main__':
    main()


'''
Created on Aug 2, 2020

@author: vladislavkargin
'''
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import root_scalar
from scipy.optimize import minimize_scalar

import rgs.mndrpy.Meander as mr

''' A couple of functions and plots for the paper about meanders '''

def entropy(x):
    ''' calculates the (binary) entropy function'''
    H = -x * np.log2(x) - (1 - x) * np.log2(1 - x)
    return H

def growthClusterRate(x, meanderRate = 8):
    ''' This function is supposed to calculate the growth in the 
    expected number of cluster-like cycles in a random meander.
    arguments:
    x -- the ratio of ringlets inside the cycle support to the half-length of
    the cycle. Can be between 0 and infty. Can be an array
    meanderRate (M) is the parameter for the growth of meanders (number of meanders of 
    half-length n is = M^n) It is known that it is between 7  a 13.
    returns:
    the growth rate for each choice of x
    '''
    
    R = 16/meanderRate
    #This is unfortunately a wrong expression
    y = - np.log2(R) - 4 * x + entropy(1/(1 + 2 * x)) * (2 + x) 
    
    return y

def maxGrowth(M, bounds = (0.001, 2), precision = 0.001):
    ''' finds the maximum growth rate in the expected 
    number of cluster-like cycles 
    arguments: M - the growth rate in the number of proper meanders 
    (M_n = M^n + error term)
    returns:
     maxG - the maximum growth where the maximum is over ratio of the ringlets 
    to the size of the proper meander
    maximizer - the maximizing ratio of the number of ringlets to the
    size of the meander'''
    x = np.array(np.arange(bounds[0], bounds[1], precision))
    y = growthClusterRate(x, meanderRate = M)
    maxG = np.max(y)
    maximizer = x[np.argmax(y)]
    return maxG, maximizer
    
def findBoundGrowth():    
    ''' finds the upper bound on the growth in the number of proper meanders.
    It is the rate such that the growth rate in the expected number
    of cluster-like meanders is 0.
    returns: 
    criticalM - the bound on the growth rate in the number of proper meanders.
    '''
    criticalM = root_scalar(lambda M: maxGrowth(M)[0], bracket=(7,9))
    return criticalM.root


def upperBound(x):
    '''
    This is a different realization of the same method to bound the growth rate 
    in the number of proper meander.
    parameter: x - ratio of the number of ringlets to the half-size of meander
    returns: the bound on the growth rate lambda for this choice of x. 
        (Here lambda = log2(M), where
        M is the other definition of the growth rate.)  
    '''   
    #this is unfortunately a mistaken expression 
    #y = 4 * (1+x) - (2 + x) * entropy(1/(1 + 2 * x)) 
    #The correct one is  here (note the division by x instead of multiplication
    def f(t):
        if t <= 0:
            z = 10**8
        else:
            z = 4 * (1+t) - (2 + t) * entropy(1/(1 + 2/t)) 
        return z
    try:
        y = list(map(f, x))
    except:
        y = f(x)
    return y

def plotUpperBound():
    '''
    This plots a picture for paper. It plots an upper bound on 
    the meander growth number for various alpha (ratio of number of 
    ringlets to the length of the proper meander)'''
    _, ax = plt.subplots(nrows=1, ncols=1)
    x = np.array(np.arange(0.01, 0.8, 0.01))
    z = upperBound(x)
    ax.plot(x, z)
    ax.grid(True)
    ax.set_xticks(np.arange(0,0.8,0.05))
    ax.tick_params(labelrotation = 90)
    ax.set_xlabel("alpha")
    ax.set_title("An upper bound on the (logarithmic) meander growth rate")
 
def calcCriticalM():   
    #finds the optimal bound on the meander rate M
    maxG = minimize_scalar(upperBound,bounds = (0.01, 1), method = "Brent")
    print(maxG)   
    rateGamma = maxG.fun
    rateM = 2 ** rateGamma  
    print("rate M = ", rateM)
    return rateM

def plotPictureQuasiCluster():
    ''' This is a picture of meanders that have one cycle of half-length 2
    and 2 ringlets'''
    _, ((ax1, ax2, ax3, ax4, ax5), (ax6, ax7, ax8, ax9, ax10),
        (ax11, ax12, ax13, ax14, ax15)) = plt.subplots(nrows=3, ncols=5)
        
    
    mndr = mr.Meander(None, None, uprngArray = [1,0,3,2,7,6,5,4], 
                        dprngArray = [1,0,3,2,5,4,7,6])
    mndr.draw(ax1)
    ax1.set_xticks([])
    ax1.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [1,0,7,4,3,6,5,2], 
                        dprngArray = [1,0,5,4,3,2,7,6])
    mndr.draw(ax2)
    ax2.set_xticks([])
    ax2.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [1,0,7,6,5,4,3,2], 
                        dprngArray = [1,0,3,2,5,4,7,6])
    mndr.draw(ax3)
    ax3.set_xticks([])
    ax3.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [1,0,7,4,3,6,5,2], 
                        dprngArray = [1,0,3,2,7,6,5,4])
    mndr.draw(ax4)
    ax4.set_xticks([])
    ax4.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [1,0,5,4,3,2,7,6], 
                        dprngArray = [1,0,3,2,5,4,7,6])
    mndr.draw(ax5)
    ax5.set_xticks([])
    ax5.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [7,2,1,4,3,6,5,0], 
                        dprngArray = [5,2,1,4,3,0,7,6])

    mndr.draw(ax6)
    ax6.set_xticks([])
    ax6.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [7,2,1,6,5,4,3,0], 
                        dprngArray = [3,2,1,0,5,4,7,6])
    mndr.draw(ax7)
    ax7.set_xticks([])
    ax7.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [7,2,1,4,3,6,5,0], 
                        dprngArray = [3,2,1,0,7,6,5,4])
    mndr.draw(ax8)
    ax8.set_xticks([])
    ax8.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [5,2,1,4,3,0,7,6], 
                        dprngArray = [3,2,1,0,5,4,7,6])
    mndr.draw(ax9)
    ax9.set_xticks([])
    ax9.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [7,6,3,2,5,4,1,0], 
                        dprngArray = [1,0,3,2,5,4,7,6])
    mndr.draw(ax10)
    ax10.set_xticks([])
    ax10.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [7,4,3,2,1,6,5,0], 
                        dprngArray = [1,0,3,2,7,6,5,4])
    mndr.draw(ax11)
    ax11.set_xticks([])
    ax11.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [7,2,1,4,3,6,5,0], 
                        dprngArray = [1,0,7,4,3,6,5,2])
    mndr = mr.Meander(None, None, uprngArray = [5,4,3,2,1,0,7,6], 
                        dprngArray = [1,0,3,2,5,4,7,6])
    mndr.draw(ax12)
    ax12.set_xticks([])
    ax12.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [7,2,1,4,3,6,5,0], 
                        dprngArray = [1,0,7,4,3,6,5,2])
    mndr.draw(ax13)
    ax13.set_xticks([])
    ax13.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [5,2,1,4,3,0,7,6], 
                        dprngArray = [1,0,5,4,3,2,7,6])
    mndr.draw(ax14)
    ax14.set_xticks([])
    ax14.set_yticks([])
    mndr = mr.Meander(None, None, uprngArray = [3,2,1,0,5,4,7,6], 
                        dprngArray = [1,0,3,2,5,4,7,6])
    mndr.draw(ax15)
    ax15.set_xticks([])
    ax15.set_yticks([])
    

#meander numbers from a paper by Jensen and Guttmann [up to M_{22}] 
#they coincide with numbers in the main entry of OEIS - which stops at
#this number
Mnumbers = [1, 1, 2, 8, 42, 262, 1828, 13820 ,110954, 933458, 8152860, 
            73424650, 678390116, 6405031050, 61606881612, 602188541928, 
            5969806669034, 59923200729046, 608188709574124, 6234277838531806,
            64477712119584604, 672265814872772972, 7060941974458061392, 
            74661728661167809752, 794337831754564188184]


#Meander numbers from a note in OEIS. The discrepancy with the previous 
#starts at M24
Mnumbers2 = [1, 1, 2, 8, 42, 262, 1828, 13820 ,110954, 933458, 8152860, 
            73424650, 678390116, 6405031050, 61606881612, 602188541928, 
            5969806669034, 59923200729046, 608188709574124, 6234277838531806,
            64477712119584604, 672265814872772972, 7060941974458061392,
            74661728661167809752, 794337831754570367812, 8499066628515413229282,
            91412898898828176826244, 987975910996038555989486, 
            10726008363361842734385644]

#Number of irreducible meanders
IrrNumbers = [1, 1, 2, 8, 46, 322, 2546, 21870, 199494, 1904624, 
              18846714, 191955370, 2002141126, 21303422480, 
              230553207346, 2531848587534, 28159614749270, 
              316713536035464, 3597509926531778, 41225699113145888, 
              476180721050626814, 5539597373695447322, 
              64863295574835126394, 763984568163192551672, 
              9047263176444565467566, 107672779595584140350702]

'''
Main Calculations
'''
if __name__ == '__main__':
    
    
    '''
    This seems to be wrong because of the mistake in growth cluster rate
    M = 7.467
    x = np.array(np.arange(0.01, 0.8, 0.01))
    y = growthClusterRate(x, meanderRate = M)
    
    fig1, ax1 = plt.subplots(nrows=1, ncols=1)
    ax1.plot(x, y)
    ax1.grid(True)
    
    (maxG, maximizer) = maxGrowth(M)
    print("Maximum growth in cluster probability is ", maxG)
    print("Reached at ", maximizer)     
    criticalM = findBoundGrowth() 
    print("Upper bound on the growth rate in the number of proper meanders:")
    print(criticalM)
    '''
    
    plotUpperBound()
    calcCriticalM()

    k = 24
    print(Mnumbers[k])
    print(Mnumbers[k]**(1/k))
    
    '''
    lengths = np.arange(len(Mnumbers))
    rateM = calcCriticalM()
    n = len(Mnumbers)
    prediction = rateM ** np.arange(n)
    z = np.array(Mnumbers)/prediction
    plt.figure()
    plt.plot(z)
    plt.yscale("log")
    #plt.xscale("log")
    '''
    
    plotPictureQuasiCluster()
    
    plt.figure()
    plt.plot(np.array(Mnumbers)/np.array(IrrNumbers[:25]))
    plt.yscale('log')
    plt.grid(True)
    
    R = (IrrNumbers[24]/Mnumbers[24])**(1/24)
    print(13.39408 - R)
    
    plt.show()
    

    
    pass


'''
Created on Apr 25, 2020

@author: vladislavkargin
'''

class Graph(object):
    '''
    this is a class for a simple graph object
    TODO: This is not yet complete since I don't actually use it 
    '''

    #Creates a graph with v vertices and no edges
    def __init__(self, V = 0):
        self.V = V
        self.E = 0
        self.adj = []; #list of adjacency lists for vertices
    '''
    Returns a string representation of this graph.
    
    return the number of vertices V, followed by the number
    of edges E, followed by the V adjacency lists
    '''
    def __str__(self):
        s = "V = " + str(self.V) + ', E = ' + str(self.E) + '\n'
        for i in range(self.V):
            s = s + str(i) + " -> " + str(self.adj[i]) + '\n'
        return s    
    
    '''
    calculates the degree of vertex v
    '''
    def degree(self, v):
        d = len(self.adj[v])
        return d
    
        # Adds the undirected edge u-v to this graph.    
    def addEdge(self, u, v): 
        self.E = self.E + 1
        self.adj[u].append(v)
        self.adj[v].append(u) 


'''
Created on Aug 14, 2020

@author: vladislavkargin
'''
import rgs.mndrpy.Meander as Meander
import rgs.mndrpy.Utility as ut
import rgs.mndrpy.Pairing as pg


import numpy as np
import itertools 
import matplotlib.pyplot as plt
#from progress.bar import Bar

def enumPropMeanders(n):
    ''' enumerates proper Meanders on [2n]. (Obviously, n cannot be too large.)
    '''
    meanders = set([])
    for uPrng in ut.genPairings(n):
        for dPrng in ut.genPairings(n):
            mndr = Meander.Meander(uPrng, dPrng)
            if mndr.isProper():
                meanders.add(mndr)
    return meanders       

def enumIrrMeanders(n):
    ''' enumerates irreducible Meanders on [2n]. (Obviously, n cannot be too large.)
    '''
    meanders = set([])
    for uPrng in ut.genPairings(n):
        for dPrng in ut.genPairings(n):
            mndr = Meander.Meander(uPrng, dPrng)
            if mndr.isIrreducible():
                meanders.add(mndr)
    return meanders     

def rotationNumber(cycle):
    '''calculate the rotation number for a cycle
    The input is a permuted list of numbers from 0 to 2n - 1 that starts with 0.
    It represents a cycle in the plane (a meander with self-intersections allowed)
    For example (0 2 1 3) represents a cycle formed by arcs (0 2) and (1 3) in 
    the upper-halfplane and arcs (2 1) (3 0) in the lower half-plane
    The function returns the rotation number for this cycle. For example, for this
    cycle the rotation number is 2. For non-crossing meanders the rotation number
    is always 0. '''
    if len(cycle)%2 != 0:
        print("Rotation Number says: The length of the cycle must be even!")
        return
    n = int(len(cycle)/2)
    #print(n)
    rn = 0
    for i in range(2*n):
        if cycle[(i + 1)%(2*n)] > cycle[i]:
            rn = rn + (-1)**(i%2)
        else:
            rn = rn - (-1)**(i%2)
    return int(rn/2) 
    
def rnDistribution(n):
    ''' Calculates the distribution of rotation numbers for all cycles of length 
    2n '''
    counts = {}
    for cycle in itertools.permutations(range(1,2 * n)):
        cycle2 = [0] + list(cycle)
        #print(cycle2)
        rn = rotationNumber(cycle2)
        if rn in counts:
            counts[rn] = counts[rn] + 1
        else:
            counts[rn] = 1
    return counts

def cycleLenDistr(n):
    ''' calculates the distribution of meander systems over cycle lengths 
    (exactly). Returns smth like [8 12 5] which means 8 meander systems with 
    only one cycle (proper meanders), 12 - with two cycles and 5 with 3 cycles'''
    
    D = [0] * n
    for mndr in ut.allMeanders(n):
        cycles, _ = mndr.findCycles()
        c = len(cycles)
        D[c - 1] = D[c - 1] + 1
    return D
    
def areaDyckPathDistr(n):
    ''' calculate the distribution of all Dyck Paths by area (Carlitz'
    variant of q-Catalan numbers)'''
    
    D = [0] * (n*(n-1)//2 + 1) #this is possible values for area
    for prng in ut.allPairings(n):
        path = pg.prng2path(prng)
        area = ut.areaDyckPath(path)
        D[area] += 1
    return D
        
def areaProperMeanderDistr(n):   
    ''' calculate the distribution of proper meanders by area '''
    meanders = enumPropMeanders(n)
    D = [0] * (n*(n-1) + 1) #this is possible values for area
    for mndr in meanders:
        area = ut.areaMeander(mndr)
        D[area] += 1
    return D
    
'''
For testing methods
'''
def main():
    #n = 7
    n = 4
    meanders = enumPropMeanders(n)
    print(len(meanders))
    #print(list(map(str,meanders)))
    
    meanders = enumIrrMeanders(n)
    print(len(meanders))
    #print(list(map(str,meanders)))
    
    print(list(itertools.permutations(range(1,n))))
    cycle = [0, 1, 5, 2, 4, 3]
    mndr = ut.makeMeanderFromCycle(cycle)
    print(mndr)
    mndr.draw()
    rn = rotationNumber(cycle)
    print("Rotation number = ", rn) 
    
    cycle = [0, 5, 1, 4, 2, 3]
    mndr = ut.makeMeanderFromCycle(cycle)
    print(mndr)
    mndr.draw()
    rn = rotationNumber(cycle)
    print("Rotation number = ", rn) 
    
    
    n = 5
    counts = rnDistribution(n) #6 is max that it can do in reasonable time
    print(counts)
    
    '''calculate cycle length distribution for meander systems
    and study its properties. A version of this code is available in PlayMeander
    Notebook on Colab.''' 
    '''
    n = 5 #n = 9 takes a rather long although reasonable time. 
    D = cycleLenDistr(n)
    print(D)
    
        
    rts = np.roots(D)
    print(rts)
    X = [z.real for z in rts]
    Y = [z.imag for z in rts]
    plt.figure()
    plt.scatter(X, Y, color = "red")
    
    print("Value of cycle distribution polynomial at -1 is ", 
          np.polyval(np.flip(D),-1))   
    '''
    
    '''now let us study the properties of the distribution 
    of Dyck paths by area. 
    A version of this code was put on the PlayMeander notebook
    on Colab.'''
    ''' 
    n = 5
    D = areaDyckPathDistr(n)
    print("Distribution by area: ", D)
    rts = np.roots(D)
    print(rts)
    X = [z.real for z in rts]
    Y = [z.imag for z in rts]
    plt.figure()
    plt.scatter(X, Y, color = "red")
    ''' 
    
    #Let us look on distribution of proper meanders by area
    n = 5 #n = 6 is already pretty tough 
    #(possible but takes some time to complete)
    meanders = enumPropMeanders(n)
    print("Number of proper meanders of size ", n, 
          " is ", len(meanders))
    D = areaProperMeanderDistr(n)
    print(D)
    
    plt.show()

if __name__ == '__main__':
    main()